from apps.monero.xmr.serialize.base_types import (
    BoolType,
    SizeT,
    UInt8,
    UInt32,
    UInt64,
    UVarintType,
)
from apps.monero.xmr.serialize.message_types import ContainerType, MessageType
from apps.monero.xmr.serialize_messages.addr import SubaddressIndex
from apps.monero.xmr.serialize_messages.base import ECKey, ECPublicKey, Hash, KeyImage
from apps.monero.xmr.serialize_messages.tx_dest_entry import TxDestinationEntry
from apps.monero.xmr.serialize_messages.tx_full import RctSig
from apps.monero.xmr.serialize_messages.tx_prefix import TransactionPrefix
from apps.monero.xmr.serialize_messages.tx_src_entry import TxSourceEntry


class MultisigOut(MessageType):
    @staticmethod
    def f_specs():
        return (("c", ContainerType, ECKey),)


class MultisigLR(MessageType):
    __slots__ = ("L", "R")

    @staticmethod
    def f_specs():
        return (("L", ECKey), ("R", ECKey))


class MultisigInfo(MessageType):
    __slots__ = ("signer", "LR", "partial_key_images")

    @staticmethod
    def f_specs():
        return (
            ("signer", ECPublicKey),
            ("LR", ContainerType, MultisigLR),
            ("partial_key_images", ContainerType, KeyImage),
        )


class MultisigStruct(MessageType):
    __slots__ = ("sigs", "ignore", "used_L", "signing_keys", "msout")

    @staticmethod
    def f_specs():
        return (
            ("sigs", RctSig),
            ("ignore", ECPublicKey),
            ("used_L", ContainerType, ECKey),
            ("signing_keys", ContainerType, ECPublicKey),
            ("msout", MultisigOut),
        )


class TransferDetails(MessageType):
    @staticmethod
    def f_specs():
        return (
            ("m_block_height", UInt64),
            ("m_tx", TransactionPrefix),
            ("m_txid", Hash),
            ("m_internal_output_index", SizeT),
            ("m_global_output_index", UInt64),
            ("m_spent", BoolType),
            ("m_spent_height", UInt64),
            ("m_key_image", KeyImage),
            ("m_mask", ECKey),
            ("m_amount", UInt64),
            ("m_rct", BoolType),
            ("m_key_image_known", BoolType),
            ("m_pk_index", SizeT),
            ("m_subaddr_index", SubaddressIndex),
            ("m_key_image_partial", BoolType),
            ("m_multisig_k", ContainerType, ECKey),
            ("m_multisig_info", ContainerType, MultisigInfo),
        )


class TxConstructionData(MessageType):
    @staticmethod
    def f_specs():
        return (
            ("sources", ContainerType, TxSourceEntry),
            ("change_dts", TxDestinationEntry),
            ("splitted_dsts", ContainerType, TxDestinationEntry),
            ("selected_transfers", ContainerType, SizeT),
            ("extra", ContainerType, UInt8),
            ("unlock_time", UInt64),
            ("use_rct", BoolType),
            ("dests", ContainerType, TxDestinationEntry),
            ("subaddr_account", UInt32),
            ("subaddr_indices", ContainerType, UVarintType),  # original: x.UInt32
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import random
import base64
import unittest
import pkg_resources

import asyncio
import aiounittest

from .. import xmrserialize as x
from .. import xmrtypes as xmr


__author__ = 'dusanklinec'


class XmrTestData(object):
    """Simple tests data generator"""

    def __init__(self, *args, **kwargs):
        super(XmrTestData, self).__init__()
        self.ec_offset = 0

    def reset(self):
        self.ec_offset = 0

    def generate_ec_key(self, use_offset=True):
        """
        Returns test EC key, 32 element byte array
        :param use_offset:
        :return:
        """
        offset = 0
        if use_offset:
            offset = self.ec_offset
            self.ec_offset += 1

        return bytearray(range(offset, offset+32))

    def gen_transaction_prefix(self):
        """
        Returns test transaction prefix
        :return:
        """
        vin = [
            xmr.TxinToKey(amount=123, key_offsets=[1, 2, 3, 2 ** 76], k_image=bytearray(range(32))),
            xmr.TxinToKey(amount=456, key_offsets=[9, 8, 7, 6], k_image=bytearray(range(32, 64))),
            xmr.TxinGen(height=99)
        ]

        vout = [
            xmr.TxOut(amount=11, target=xmr.TxoutToKey(key=bytearray(range(32)))),
            xmr.TxOut(amount=34, target=xmr.TxoutToKey(key=bytearray(range(64, 96)))),
        ]

        msg = xmr.TransactionPrefix(version=2, unlock_time=10, vin=vin, vout=vout, extra=list(range(31)))
        return msg

    def gen_borosig(self):
        """
        Returns a BoroSig message
        :return:
        """
        ee = self.generate_ec_key()
        s0 = [self.generate_ec_key() for _ in range(64)]
        s1 = [self.generate_ec_key() for _ in range(64)]
        msg = xmr.BoroSig(s0=s0, s1=s1, ee=ee)
        return msg


if __name__ == "__main__":
    unittest.main()  # pragma: no cover





from apps.monero.xmr.serialize.erefs import eref
from apps.monero.xmr.serialize.message_types import ContainerType, MessageType
from apps.monero.xmr.serialize_messages.base import ECKey
from apps.monero.xmr.serialize_messages.tx_prefix import (
    TxinGen,
    TxinToKey,
    TxinToScript,
    TxinToScriptHash,
)


class Signature(MessageType):
    __slots__ = ("c", "r")

    @staticmethod
    def f_specs():
        return (("c", ECKey), ("r", ECKey))

    async def serialize_archive(self, ar):
        ar.field(eref(self, "c"), ECKey)
        ar.field(eref(self, "r"), ECKey)
        return self


class SignatureArray(ContainerType):
    FIX_SIZE = 0
    ELEM_TYPE = Signature


def get_signature_size(msg):
    """
    Returns a signature size for the input
    :param msg:
    :return:
    """
    if isinstance(msg, (TxinGen, TxinToScript, TxinToScriptHash)):
        return 0
    elif isinstance(msg, TxinToKey):
        return len(msg.key_offsets)
    else:
        raise ValueError("Unknown tx in")



_UINT_BUFFER = bytearray(1)


async def load_uint(reader, width):
    """
    Constant-width integer serialization
    :param reader:
    :param width:
    :return:
    """
    buffer = _UINT_BUFFER
    result = 0
    shift = 0
    for _ in range(width):
        await reader.areadinto(buffer)
        result += buffer[0] << shift
        shift += 8
    return result


async def dump_uint(writer, n, width):
    """
    Constant-width integer serialization
    :param writer:
    :param n:
    :param width:
    :return:
    """
    buffer = _UINT_BUFFER
    for _ in range(width):
        buffer[0] = n & 0xff
        await writer.awrite(buffer)
        n >>= 8


def uvarint_size(n):
    """
    Returns size in bytes n would occupy serialized as varint
    :param n:
    :return:
    """
    bts = 0 if n != 0 else 1
    while n:
        n >>= 7
        bts += 1
    return bts


def load_uvarint_b(buffer):
    """
    Variable int deserialization, synchronous from buffer.
    :param buffer:
    :return:
    """
    result = 0
    idx = 0
    byte = 0x80
    while byte & 0x80:
        byte = buffer[idx]
        result += (byte & 0x7F) << (7 * idx)
        idx += 1
    return result


def dump_uvarint_b(n):
    """
    Serializes uvarint to the buffer
    :param n:
    :return:
    """
    buffer = bytearray(uvarint_size(n))
    return dump_uvarint_b_into(n, buffer, 0)


def dump_uvarint_b_into(n, buffer, offset=0):
    """
    Serializes n as variable size integer to the provided buffer.
    Buffer has to ha
    :param n:
    :param buffer:
    :param offset:
    :return:
    """
    shifted = True
    while shifted:
        shifted = n >> 7
        buffer[offset] = (n & 0x7F) | (0x80 if shifted else 0x00)
        offset += 1
        n = shifted
    return buffer


def load_uint_b(buffer, width):
    """
    Loads fixed size integer from the buffer
    :param buffer:
    :return:
    """
    result = 0
    for idx in range(width):
        result += buffer[idx] << (8 * idx)
    return result


def dump_uint_b(n, width):
    """
    Serializes fixed size integer to the buffer
    :param n:
    :param width:
    :return:
    """
    buffer = bytearray(width)
    return dump_uvarint_b_into(n, buffer, 0)


def dump_uint_b_into(n, width, buffer, offset=0):
    """
    Serializes fixed size integer to the buffer
    :param n:
    :param width:
    :return:
    """
    for idx in range(width):
        buffer[idx + offset] = n & 0xff
        n >>= 8
    return buffer



#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
Monero serialization to object representation.

Note: This is not optimized for memory usage. No streaming support.
Used for debugging.
'''

import base64
import collections

from . import xmrserialize as x
from .xmrserialize import eref, get_elem, set_elem


class Archive(x.Archive):
    """
    Archive object for object binary serialization / deserialization.
    Resembles Archive API from the Monero codebase or Boost serialization archive.

    The design goal is to provide uniform API both for serialization and deserialization
    so the code is not duplicated for serialization and deserialization but the same
    for both ways in order to minimize potential bugs in the code.

    In order to use the archive for both ways we have to use so-called field references
    as we cannot directly modify given element as a parameter (value-passing) as its performed
    in C++ code. see: eref(), get_elem(), set_elem()
    """
    def __init__(self, iobj, writing=True, **kwargs):
        super().__init__(iobj, writing, **kwargs)
        self.res = collections.OrderedDict()

    async def tag(self, tag):
        """
        TODO: implement
        :param tag:
        :return:
        """

    async def begin_array(self):
        """
        Mark start of the array. Used for JSON serialization.
        TODO: implement
        :return:
        """

    async def end_array(self):
        """
        Mark end of the array. Used for JSON serialization.
        TODO: implement
        :return:
        """

    async def begin_object(self):
        """
        Mark start of the object. Used for JSON serialization.
        TODO: implement
        :return:
        """

    async def end_object(self):
        """
        Mark end of the object. Used for JSON serialization.
        TODO: implement
        :return:
        """


async def dump_blob(elem, elem_type=None):
    """
    Dumps blob message.
    Supports both blob and raw value.

    :param writer:
    :param elem:
    :param elem_type:
    :param params:
    :return:
    """
    elem_is_blob = isinstance(elem, x.BlobType)
    data = getattr(elem, x.BlobType.DATA_ATTR) if elem_is_blob else elem
    if data is None or len(data) == 0:
        return b''
    if isinstance(data, (bytes, bytearray, list)):
        return base64.b16encode(bytes(data))
    else:
        raise ValueError('Unknown blob type')


async def load_blob(elem, elem_type=None):
    """
    Loads blob from serialized object
    :param elem:
    :return:
    """
    if elem is None:
        return b''
    return bytearray(base64.b16decode(elem))


async def dump_container(obj, container, container_type, params=None, field_archiver=None):
    """
    Serializes container as popo

    :param obj:
    :param container:
    :param container_type:
    :param params:
    :param field_archiver:
    :return:
    """
    field_archiver = field_archiver if field_archiver else dump_field
    elem_type = params[0] if params else None
    if elem_type is None:
        elem_type = container_type.ELEM_TYPE

    obj = [] if obj is None else get_elem(obj)
    if container is None:
        return None
    for elem in container:
        fvalue = await field_archiver(None, elem, elem_type, params[1:] if params else None)
        obj.append(fvalue)
    return obj


async def load_container(obj, container_type, params=None, container=None, field_archiver=None):
    """
    Loads container of elements from the object representation. Supports the container ref.
    Returns loaded container.

    :param reader:
    :param container_type:
    :param params:
    :param container:
    :param field_archiver:
    :return:
    """
    field_archiver = field_archiver if field_archiver else load_field
    if obj is None:
        return None

    c_len = len(obj)
    elem_type = params[0] if params else None
    if elem_type is None:
        elem_type = container_type.ELEM_TYPE

    res = container if container else []
    for i in range(c_len):
        fvalue = await field_archiver(obj[i], elem_type,
                                      params[1:] if params else None,
                                      eref(res, i) if container else None)
        if not container:
            res.append(fvalue)
    return res


async def dump_message_field(obj, msg, field, field_archiver=None):
    """
    Dumps a message field to the object. Field is defined by the message field specification.

    :param obj:
    :param msg:
    :param field:
    :param field_archiver:
    :return:
    """
    fname, ftype, params = field[0], field[1], field[2:]
    fvalue = getattr(msg, fname, None)
    field_archiver = field_archiver if field_archiver else dump_field
    return await field_archiver(eref(obj, fname, True), fvalue, ftype, params)


async def load_message_field(obj, msg, field, field_archiver=None):
    """
    Loads message field from the object. Field is defined by the message field specification.
    Returns loaded value, supports field reference.

    :param reader:
    :param msg:
    :param field:
    :param field_archiver:
    :return:
    """
    fname, ftype, params = field[0], field[1], field[2:]
    field_archiver = field_archiver if field_archiver else load_field
    await field_archiver(obj[fname], ftype, params, eref(msg, fname))


async def dump_message(obj, msg, field_archiver=None):
    """
    Dumps message to the object.
    Returns message popo representation.

    :param obj:
    :param msg:
    :param field_archiver:
    :return:
    """
    mtype = msg.__class__
    fields = mtype.f_specs()

    obj = collections.OrderedDict() if obj is None else get_elem(obj)
    for field in fields:
        await dump_message_field(obj, msg=msg, field=field, field_archiver=field_archiver)
    return obj


async def load_message(obj, msg_type, msg=None, field_archiver=None):
    """
    Loads message if the given type from the object.
    Supports reading directly to existing message.

    :param obj:
    :param msg_type:
    :param msg:
    :param field_archiver:
    :return:
    """
    msg = msg_type() if msg is None else msg

    fields = msg_type.f_specs() if msg_type else msg.__class__.f_specs()
    for field in fields:
        await load_message_field(obj, msg, field, field_archiver=field_archiver)

    return msg


async def dump_variant(obj, elem, elem_type=None, params=None, field_archiver=None):
    """
    Transform variant to the popo object representation.

    :param obj:
    :param elem:
    :param elem_type:
    :param params:
    :param field_archiver:
    :return:
    """
    field_archiver = field_archiver if field_archiver else dump_field
    if isinstance(elem, x.VariantType) or elem_type.WRAPS_VALUE:
        return {
            elem.variant_elem: await field_archiver(None, getattr(elem, elem.variant_elem), elem.variant_elem_type)
        }

    else:
        fdef = elem_type.find_fdef(elem_type.f_specs(), elem)
        return {
            fdef[0]: await field_archiver(None, elem, fdef[1])
        }


async def load_variant(obj, elem, elem_type=None, params=None, field_archiver=None, wrapped=None):
    """
    Loads variant from the obj representation
    :param obj:
    :param elem:
    :param elem_type:
    :param params:
    :param field_archiver:
    :param wrapped:
    :return:
    """
    field_archiver = field_archiver if field_archiver else load_field
    is_wrapped = elem_type.WRAPS_VALUE if wrapped is None else wrapped

    if is_wrapped:
        elem = elem_type() if elem is None else elem

    fname = list(obj.keys())[0]
    for field in elem_type.f_specs():
        if field[0] != fname:
            continue

        fvalue = await field_archiver(obj[fname], field[1], field[2:], elem if not is_wrapped else None)
        if is_wrapped:
            elem.set_variant(field[0], fvalue)

        return elem if is_wrapped else fvalue
    raise ValueError('Unknown tag: %s' % fname)


async def dump_field(obj, elem, elem_type, params=None):
    """
    Dumps generic field to the popo object representation, according to the element specification.
    General multiplexer.

    :param obj:
    :param elem:
    :param elem_type:
    :param params:
    :return:
    """
    if isinstance(elem, (int, bool)) or issubclass(elem_type, x.UVarintType) or issubclass(elem_type, x.IntType):
        return set_elem(obj, elem)

    elif issubclass(elem_type, x.BlobType) or isinstance(obj, bytes) or isinstance(obj, bytearray):
        return set_elem(obj, await dump_blob(elem))

    elif issubclass(elem_type, x.UnicodeType) or isinstance(elem, str):
        return set_elem(obj, elem)

    elif issubclass(elem_type, x.VariantType):
        return set_elem(obj, await dump_variant(None, elem, elem_type, params))

    elif issubclass(elem_type, x.ContainerType):  # container ~ simple list
        return set_elem(obj, await dump_container(None, elem, elem_type, params))

    elif issubclass(elem_type, x.MessageType):
        return set_elem(obj, await dump_message(None, elem))

    else:
        raise TypeError


async def load_field(obj, elem_type, params=None, elem=None):
    """
    Loads a field from the reader, based on the field type specification. Demultiplexer.

    :param obj:
    :param elem_type:
    :param params:
    :param elem:
    :return:
    """
    if issubclass(elem_type, x.UVarintType) or issubclass(elem_type, x.IntType) or isinstance(obj, (int, bool)):
        return set_elem(elem, obj)

    elif issubclass(elem_type, x.BlobType):
        fvalue = await load_blob(obj, elem_type)
        return set_elem(elem, fvalue)

    elif issubclass(elem_type, x.UnicodeType) or isinstance(elem, str):
        return set_elem(elem, obj)

    elif issubclass(elem_type, x.VariantType):
        fvalue = await load_variant(obj, elem=get_elem(elem), elem_type=elem_type, params=params)
        return set_elem(elem, fvalue)

    elif issubclass(elem_type, x.ContainerType):  # container ~ simple list
        fvalue = await load_container(obj, elem_type, params=params, container=get_elem(elem))
        return set_elem(elem, fvalue)

    elif issubclass(elem_type, x.MessageType):
        fvalue = await load_message(obj, msg_type=elem_type, msg=get_elem(elem))
        return set_elem(elem, fvalue)

    else:
        raise TypeError





#!/usr/bin/env python
# -*- coding: utf-8 -*-
import random
import unittest
import binascii
import json
import os
import pkg_resources

import asyncio
import aiounittest

from .test_data import XmrTestData
from .. import xmrserialize as x
from .. import xmrtypes as xmr
from .. import xmrboost as xmrb
from .. import xmrobj as xmro
from .. import xmrjson as xmrjs


__author__ = 'dusanklinec'


class XmrBoostTest(aiounittest.AsyncTestCase):
    """Simple tests"""

    def __init__(self, *args, **kwargs):
        super(XmrBoostTest, self).__init__(*args, **kwargs)
        self.test_data = XmrTestData()

    def setUp(self):
            self.test_data.reset()

    async def test_simple_msg(self):
        """
        TxinGen
        :return:
        """
        data_hex = b'011673657269616c697a6174696f6e3a3a61726368697665000000000134'
        data_bin = binascii.unhexlify(data_hex)
        reader = x.MemoryReaderWriter(bytearray(data_bin))
        ar = xmrb.Archive(reader, False)

        msg = xmr.TxinGen()
        await ar.root_message(msg)
        self.assertEqual(msg.height, 0x34)

        writer = x.MemoryReaderWriter()
        ar2 = xmrb.Archive(writer, True)
        await ar2.root()
        await ar2.message(msg)
        self.assertEqual(data_bin, bytearray(writer.get_buffer()))

    async def test_ctkey_msg(self):
        """
        CtKey
        :return:
        """
        data_hex = b'011673657269616c697a6174696f6e3a3a617263686976650000000000000120000000000000000000000000000000000000000000000000000000000000000001200000000000000000000000000000000000000000000000000000000000000000'
        data_bin = binascii.unhexlify(data_hex)
        reader = x.MemoryReaderWriter(bytearray(data_bin))
        ar = xmrb.Archive(reader, False)

        msg = xmr.CtKey()
        await ar.root_message(msg)

        writer = x.MemoryReaderWriter()
        ar2 = xmrb.Archive(writer, True)
        await ar2.root()
        await ar2.message(msg)
        self.assertEqual(data_bin, bytearray(writer.get_buffer()))

    async def test_destination_entries(self):
        """
        Tx destinations, two records, versioning test.
        :return:
        """
        data_hex = b'011673657269616c697a6174696f6e3a3a6172636869766500000001010144000000000120cc000000000000000000000000000000000000000000000000000000000000ee0120aa000000000000000000000000000000000000000000000000000000000000dd01010122012011000000000000000000000000000000000000000000000000000000000000ee012033000000000000000000000000000000000000000000000000000000000000dd00'
        data_bin = binascii.unhexlify(data_hex)
        reader = x.MemoryReaderWriter(bytearray(data_bin))
        ar = xmrb.Archive(reader, False)

        msg = xmr.TxDestinationEntry()
        await ar.root_message(msg)

        self.assertEqual(msg.amount, 0x44)
        self.assertEqual(msg.is_subaddress, 1)
        self.assertEqual(msg.addr.m_spend_public_key, bytearray(b'\xcc' + (b'\x00' * 30) + b'\xee'))
        self.assertEqual(msg.addr.m_view_public_key, bytearray(b'\xaa' + (b'\x00' * 30) + b'\xdd'))

        msg2 = xmr.TxDestinationEntry()
        await ar.message(msg2)

        self.assertEqual(msg2.amount, 0x22)
        self.assertEqual(msg2.is_subaddress, 0)
        self.assertEqual(msg2.addr.m_spend_public_key, bytearray(b'\x11' + (b'\x00' * 30) + b'\xee'))
        self.assertEqual(msg2.addr.m_view_public_key, bytearray(b'\x33' + (b'\x00' * 30) + b'\xdd'))

        writer = x.MemoryReaderWriter()
        ar2 = xmrb.Archive(writer, True)
        await ar2.root()
        await ar2.message(msg)
        await ar2.message(msg2)
        self.assertEqual(data_bin, bytearray(writer.get_buffer()))

    async def test_tx_prefix(self):
        """
        Transaction prefix
        :return:
        """
        data_hex = pkg_resources.resource_string(__name__, os.path.join('data', 'tx_prefix_01.txt'))
        data_bin = binascii.unhexlify(data_hex)
        reader = x.MemoryReaderWriter(bytearray(data_bin))
        ar = xmrb.Archive(reader, False)

        msg = xmr.TransactionPrefix()
        await ar.root()
        await ar.message(msg)
        self.assertEqual(msg.version, 2)
        self.assertEqual(msg.unlock_time, 0)
        self.assertEqual(len(msg.vin), 1)
        self.assertEqual(len(msg.vout), 2)
        self.assertEqual(len(msg.extra), 44)
        self.assertEqual(msg.vin[0].amount, 0)
        self.assertEqual(len(msg.vin[0].key_offsets), 7)
        self.assertEqual(msg.vin[0].key_offsets[6], 56)
        self.assertEqual(msg.vout[0].amount, 0)
        self.assertEqual(msg.vout[1].amount, 0)
        self.assertEqual(msg.vout[0].target.key,
                         binascii.unhexlify(b'8829dedf62f2c7e6b21556daed786143fb4c1c3ff95bde568b82860229c44b17'))
        self.assertEqual(msg.vout[1].target.key,
                         binascii.unhexlify(b'98d2c70dcef9fc62378ed2d49301a800f90cd8c9423c2b08dfef5f9635267978'))
        self.assertEqual(msg.extra, [2, 9, 1, 47, 52, 19, 57, 144, 50, 174, 234, 1, 115, 225, 189, 191, 247, 206, 185, 246, 32, 137, 241, 205, 17, 99, 170, 72, 226, 34, 108, 233, 7, 125, 90, 190, 17, 206, 149, 202, 216, 243, 78, 13])

        writer = x.MemoryReaderWriter()
        ar2 = xmrb.Archive(writer, True)
        await ar2.root()
        await ar2.message(msg)
        self.assertEqual(data_bin, bytearray(writer.get_buffer()))

    async def test_tx(self):
        """
        Full transaction
        :return:
        """
        data_hex = pkg_resources.resource_string(__name__, os.path.join('data', 'tx_01.txt'))
        data_bin = binascii.unhexlify(data_hex)
        reader = x.MemoryReaderWriter(bytearray(data_bin))
        ar = xmrb.Archive(reader, False)

        msg = xmr.Transaction()
        await ar.root()
        await ar.message(msg)
        self.assertEqual(msg.version, 2)
        self.assertEqual(msg.unlock_time, 0)
        self.assertEqual(msg.rct_signatures.type, 1)
        self.assertEqual(msg.rct_signatures.txnFee, 9119110000)
        self.assertEqual(msg.rct_signatures.outPk[1].mask,
                         binascii.unhexlify(b'4908ea58bad7ed52bd70cc201eaae6280d833226bce333b16672c30ad91f9b43'))
        self.assertEqual(len(msg.rct_signatures.p.rangeSigs), 2)
        self.assertEqual(len(msg.rct_signatures.p.MGs), 1)
        self.assertEqual(len(msg.rct_signatures.p.MGs[0].ss), 7)
        self.assertEqual(msg.rct_signatures.p.MGs[0].cc,
                         binascii.unhexlify(b'23237696b7cfd0fe1159cfedd00a8acc138ef91f2f249e6289e1f14cb58bce06'))

        writer = x.MemoryReaderWriter()
        ar2 = xmrb.Archive(writer, True)
        await ar2.root()
        await ar2.message(msg)
        self.assertEqual(data_bin, bytearray(writer.get_buffer()))

    async def test_tx_metadata(self):
        """
        Tx metadata produced by watch-only wallet
        :return:
        """
        data_hex = pkg_resources.resource_string(__name__, os.path.join('data', 'tx_metadata_01.txt'))
        data_bin = binascii.unhexlify(data_hex)
        reader = x.MemoryReaderWriter(bytearray(data_bin))
        ar = xmrb.Archive(reader, False)

        msg = xmr.PendingTransaction()
        await ar.root()
        await ar.message(msg)

        self.assertEqual(msg.tx_key,
                         binascii.unhexlify(b'a0a50810dbc38101a79525823428b500ac936dfea613c73b4864f7260ff26a0b'))
        self.assertEqual(msg.change_dts.amount, 99972803971000)
        self.assertEqual(msg.fee, 9119110000)
        self.assertEqual(msg.construction_data.use_rct, 1)
        self.assertEqual(len(msg.construction_data.extra), 44)
        self.assertEqual(len(msg.construction_data.sources), 1)
        self.assertEqual(msg.construction_data.sources[0].amount, 100000000000000)
        self.assertEqual(msg.construction_data.sources[0].mask,
                         binascii.unhexlify(b'2dea8778cf4e89a7f32b5659d674d44795a370a00f79ee9b2ea37c1fcb005c0d'))
        self.assertEqual(len(msg.construction_data.sources[0].outputs), 7)
        self.assertEqual(msg.construction_data.sources[0].outputs[6][0], 1727)
        self.assertEqual(msg.construction_data.sources[0].outputs[6][1].mask,
                         binascii.unhexlify(b'2eeec82a970bfa54c35b0b740f6fb0585de14818e3c6dceed75c76fe69e3e449'))

        self.assertEqual(len(msg.construction_data.dests), 1)
        self.assertEqual(len(msg.construction_data.splitted_dsts), 2)
        self.assertEqual(msg.construction_data.splitted_dsts[0].amount, 18076919000)
        self.assertEqual(msg.construction_data.splitted_dsts[1].amount, 99972803971000)
        self.assertEqual(len(msg.construction_data.subaddr_indices), 1)

        vers = xmrb.VersionSetting()
        vers.set(xmr.TxConstructionData, 2)

        writer = x.MemoryReaderWriter()
        ar2 = xmrb.Archive(writer, True, vers)
        await ar2.root()
        await ar2.message(msg)
        self.assertEqual(data_bin, bytearray(writer.get_buffer()))

        msg.construction_data.use_bulletproofs = False
        writer = x.MemoryReaderWriter()
        ar2 = xmrb.Archive(writer, True)
        await ar2.root()
        await ar2.message(msg)

    async def test_tx_unsigned(self):
        unsigned_tx_c = pkg_resources.resource_string(__name__, os.path.join('data', 'tx_unsigned_01.txt'))
        unsigned_tx = binascii.unhexlify(unsigned_tx_c)

        reader = x.MemoryReaderWriter(bytearray(unsigned_tx))
        ar = xmrb.Archive(reader, False)

        msg = xmr.UnsignedTxSet()
        await ar.root()
        await ar.message(msg)
        self.assertEqual(len(msg.transfers), 2)
        self.assertEqual(msg.transfers[0].m_block_height, 1998)
        self.assertEqual(msg.transfers[0].m_global_output_index, 701)
        self.assertEqual(msg.transfers[1].m_block_height, 3312)
        self.assertEqual(msg.transfers[1].m_global_output_index, 2026)
        self.assertEqual(msg.transfers[1].m_amount, 1000000000000000)
        self.assertEqual(msg.transfers[1].m_mask,
                         bytearray([0x5c,0x2f,0x4b,0x93,0x26,0xd1,0xa3,0xd8,0x17,0x0d,0x1e,0x5b,0x69,0xb5,0x19,0x2c,0xba,0x9d,0x7c,0x48,0xf2,0xc7,0xc3,0xcf,0xdd,0x9d,0x1b,0xbd,0x4f,0x96,0xeb,0x00]))

        writer = x.MemoryReaderWriter()
        vers = xmrb.VersionSetting()
        vers.set(xmr.TxConstructionData, 2)

        ar2 = xmrb.Archive(writer, True, vers)
        await ar2.root()
        await ar2.message(msg)
        self.assertEqual(unsigned_tx, bytearray(writer.get_buffer()))

        for tx in msg.txes:
            tx.use_bulletproofs = False
        ar2 = xmrb.Archive(writer, True, vers)
        await ar2.root()
        await ar2.message(msg)


if __name__ == "__main__":
    unittest.main()  # pragma: no cover






#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
Minimal streaming codec for a Monero binary serialization.
Used for a binary serialization in blockchain and for hash computation for signatures.

Equivalent of BEGIN_SERIALIZE_OBJECT(), /src/serialization/serialization.h

- The wire binary format does not use tags. Structure has to be read from the binary stream
with the scheme specified in order to parse the structure.

- Heavily uses variable integer serialization - similar to the UTF8 or LZ4 number encoding.

- Supports: blob, string, integer types - variable or fixed size, containers of elements,
            variant types, messages of elements

For de-serializing (loading) types, object with `AsyncReader`
interface is required:

>>> class AsyncReader:
>>>     async def areadinto(self, buffer):
>>>         """
>>>         Reads `len(buffer)` bytes into `buffer`, or raises `EOFError`.
>>>         """

For serializing (dumping) types, object with `AsyncWriter` interface is
required:

>>> class AsyncWriter:
>>>     async def awrite(self, buffer):
>>>         """
>>>         Writes all bytes from `buffer`, or raises `EOFError`.
>>>         """
'''

import logging
import sys

from . import helpers
from .protobuf import const, load_uvarint, dump_uvarint
from .core.readwriter import MemoryReaderWriter
from .core.base_types import *
from .core.erefs import has_elem, set_elem, get_elem, ElemRefArr, ElemRefObj, eref, is_elem_ref
from .core.int_serialize import *
from .core.message_types import *
from .core.obj_helper import *


logger = logging.getLogger(__name__)


def import_def(module, name):
    if module not in sys.modules:
        if not module.startswith("monero_serialize"):
            raise ValueError("Module not allowed: %s" % module)

        logger.debug("Importing: from %s import %s", module, name)
        __import__(module, None, None, (name,), 0)

    r = getattr(sys.modules[module], name)
    return r


class Archive(object):
    """
    Archive object for object binary serialization / deserialization.
    Resembles Archive API from the Monero codebase or Boost serialization archive.

    The design goal is to provide uniform API both for serialization and deserialization
    so the code is not duplicated for serialization and deserialization but the same
    for both ways in order to minimize potential bugs in the code.

    In order to use the archive for both ways we have to use so-called field references
    as we cannot directly modify given element as a parameter (value-passing) as its performed
    in C++ code. see: eref(), get_elem(), set_elem()
    """
    def __init__(self, iobj, writing=True, **kwargs):
        self.writing = writing
        self.iobj = iobj
        self.tracker = helpers.Tracker()

    async def tag(self, tag):
        """

        :param tag:
        :return:
        """

    async def begin_array(self):
        """
        Mark start of the array. Used for JSON serialization.
        :return:
        """

    async def end_array(self):
        """
        Mark end of the array. Used for JSON serialization.
        :return:
        """

    async def begin_object(self):
        """
        Mark start of the object. Used for JSON serialization.
        :return:
        """

    async def end_object(self):
        """
        Mark end of the object. Used for JSON serialization.
        :return:
        """

    async def prepare_container(self, size, container, elem_type=None):
        """
        Prepares container for serialization
        :param size:
        :param container:
        :return:
        """
        if not self.writing:
            if container is None:
                return gen_elem_array(size, elem_type)

            fvalue = get_elem(container)
            if fvalue is None:
                fvalue = []
            fvalue += gen_elem_array(max(0, size - len(fvalue)), elem_type)
            set_elem(container, fvalue)
            return fvalue

    async def prepare_message(self, msg, msg_type):
        """
        Prepares message for serialization
        :param msg:
        :param msg_type:
        :return:
        """
        if self.writing:
            return
        return set_elem(msg, msg_type())

    async def uvarint(self, elem):
        """
        Uvarint
        :param elem:
        :return:
        """
        if self.writing:
            return await dump_uvarint(self.iobj, elem)
        else:
            return await load_uvarint(self.iobj)

    async def uint(self, elem, elem_type, params=None):
        """
        Fixed size int
        :param elem:
        :param elem_type:
        :param params:
        :return:
        """
        if self.writing:
            return await dump_uint(self.iobj, elem, elem_type.WIDTH)
        else:
            return await load_uint(self.iobj, elem_type.WIDTH)

    async def unicode_type(self, elem):
        """
        Unicode type
        :param elem:
        :return:
        """
        if self.writing:
            return await dump_unicode(self.iobj, elem)
        else:
            return await load_unicode(self.iobj)

    async def blob(self, elem=None, elem_type=None, params=None):
        """
        Loads/dumps blob
        :return:
        """
        elem_type = elem_type if elem_type else elem.__class__
        if hasattr(elem_type, "serialize_archive"):
            elem = elem_type() if elem is None else elem
            return await elem.serialize_archive(
                self, elem=elem, elem_type=elem_type, params=params
            )

        if self.writing:
            return await dump_blob(
                self.iobj, elem=elem, elem_type=elem_type, params=params
            )
        else:
            return await load_blob(
                self.iobj, elem_type=elem_type, params=params, elem=elem
            )

    async def container(self, container=None, container_type=None, params=None):
        """
        Loads/dumps container
        :return:
        """
        if hasattr(container_type, "serialize_archive"):
            container = container_type() if container is None else container
            return await container.serialize_archive(
                self, elem=container, elem_type=container_type, params=params
            )

        if self.writing:
            return await self._dump_container(
                self.iobj, container, container_type, params
            )
        else:
            return await self._load_container(
                self.iobj, container_type, params=params, container=container
            )

    async def container_size(
            self, container_len=None, container_type=None, params=None
    ):
        """
        Container size
        :param container_len:
        :param container_type:
        :param params:
        :return:
        """
        if hasattr(container_type, "serialize_archive"):
            raise ValueError("not supported")

        if self.writing:
            return await self._dump_container_size(
                self.iobj, container_len, container_type, params
            )
        else:
            raise ValueError("Not supported")

    async def container_val(self, elem, container_type, params=None):
        """
        Single cont value
        :param elem:
        :param container_type:
        :param params:
        :return:
        """
        if hasattr(container_type, "serialize_archive"):
            raise ValueError("not supported")
        if self.writing:
            return await self._dump_container_val(
                self.iobj, elem, container_type, params
            )
        else:
            raise ValueError("Not supported")

    async def tuple(self, elem=None, elem_type=None, params=None):
        """
        Loads/dumps tuple
        :return:
        """
        if hasattr(elem_type, "serialize_archive"):
            container = elem_type() if elem is None else elem
            return await container.serialize_archive(
                self, elem=elem, elem_type=elem_type, params=params
            )

        if self.writing:
            return await self._dump_tuple(self.iobj, elem, elem_type, params)
        else:
            return await self._load_tuple(
                self.iobj, elem_type, params=params, elem=elem
            )

    async def variant(self, elem=None, elem_type=None, params=None, wrapped=None):
        """
        Loads/dumps variant type
        :param elem:
        :param elem_type:
        :param params:
        :return:
        """
        elem_type = elem_type if elem_type else elem.__class__
        if hasattr(elem_type, "serialize_archive"):
            elem = elem_type() if elem is None else elem
            return await elem.serialize_archive(
                self, elem=elem, elem_type=elem_type, params=params
            )

        if self.writing:
            return await self._dump_variant(
                self.iobj,
                elem=elem,
                elem_type=elem_type if elem_type else elem.__class__,
                params=params,
            )
        else:
            return await self._load_variant(
                self.iobj,
                elem_type=elem_type if elem_type else elem.__class__,
                params=params,
                elem=elem,
                wrapped=wrapped,
            )

    async def message(self, msg, msg_type=None):
        """
        Loads/dumps message
        :param msg:
        :param msg_type:
        :return:
        """
        elem_type = msg_type if msg_type is not None else msg.__class__
        msg = elem_type() if msg is None else msg
        if hasattr(elem_type, "serialize_archive"):
            return await msg.serialize_archive(self)

        mtype = msg.__class__ if msg_type is None else msg_type
        fields = mtype.f_specs()
        if hasattr(mtype, "serialize_archive"):
            raise ValueError("Cannot directly load, has to use archive with %s" % mtype)

        for field in fields:
            await self.message_field(msg, field)
        return msg

    async def message_field(self, msg, field, fvalue=None):
        """
        Dumps/Loads message field
        :param msg:
        :param field:
        :param fvalue: explicit value for dump
        :return:
        """
        fname, ftype, params = field[0], field[1], field[2:]
        try:
            self.tracker.push_field(fname)
            if self.writing:
                await self._dump_message_field(self.iobj, msg, field, fvalue=fvalue)
            else:
                await self._load_message_field(self.iobj, msg, field)
            self.tracker.pop()

        except Exception as e:
            raise helpers.ArchiveException(e, tracker=self.tracker) from e

    async def message_fields(self, msg, fields):
        """
        Load/dump individual message fields
        :param msg:
        :param fields:
        :return:
        """
        for field in fields:
            await self.message_field(msg, field)
        return msg

    def _get_type(self, elem_type):
        # If part of our hierarchy - return the object
        if issubclass(elem_type, XmrType):
            return elem_type

        # Basic decision types
        etypes = (
            UVarintType,
            IntType,
            BlobType,
            UnicodeType,
            VariantType,
            ContainerType,
            TupleType,
            MessageType,
        )
        cname = elem_type.__name__
        for e in etypes:
            if cname == e.__name__:
                return e

        # Inferred type: need to translate it to the current
        try:
            m = elem_type.__module__
            r = import_def(m, cname)
            sub_test = issubclass(r, XmrType)
            logger.debug(
                "resolved %s, sub: %s, id_e: %s, id_mod: %s",
                r,
                sub_test,
                id(r),
                id(sys.modules[m]),
            )
            if not sub_test:
                logger.warning("resolution hierarchy broken")

            return r

        except Exception as e:
            raise ValueError(
                "Could not translate elem type: %s %s, exc: %s %s"
                % (type(elem_type), elem_type, type(e), e)
            )

    def _is_type(self, elem_type, test_type):
        return issubclass(elem_type, test_type)

    async def field(self, elem=None, elem_type=None, params=None):
        """
        Archive field
        :param elem:
        :param elem_type:
        :param params:
        :return:
        """
        elem_type = elem_type if elem_type else elem.__class__
        fvalue = None

        etype = self._get_type(elem_type)
        if self._is_type(etype, UVarintType):
            fvalue = await self.uvarint(get_elem(elem))

        elif self._is_type(etype, IntType):
            fvalue = await self.uint(
                elem=get_elem(elem), elem_type=elem_type, params=params
            )

        elif self._is_type(etype, BlobType):
            fvalue = await self.blob(
                elem=get_elem(elem), elem_type=elem_type, params=params
            )

        elif self._is_type(etype, UnicodeType):
            fvalue = await self.unicode_type(get_elem(elem))

        elif self._is_type(etype, VariantType):
            fvalue = await self.variant(
                elem=get_elem(elem), elem_type=elem_type, params=params
            )

        elif self._is_type(etype, ContainerType):  # container ~ simple list
            fvalue = await self.container(
                container=get_elem(elem), container_type=elem_type, params=params
            )

        elif self._is_type(etype, TupleType):  # tuple ~ simple list
            fvalue = await self.tuple(
                elem=get_elem(elem), elem_type=elem_type, params=params
            )

        elif self._is_type(etype, MessageType):
            fvalue = await self.message(get_elem(elem), msg_type=elem_type)

        else:
            raise TypeError(
                "unknown type: %s %s %s" % (elem_type, type(elem_type), elem)
            )

        return fvalue if self.writing else set_elem(elem, fvalue)

    async def dump_field(self, writer, elem, elem_type, params=None):
        assert self.iobj == writer
        return await self.field(elem=elem, elem_type=elem_type, params=params)

    async def load_field(self, reader, elem_type, params=None, elem=None):
        assert self.iobj == reader
        return await self.field(elem=elem, elem_type=elem_type, params=params)

    async def root(self):
        """
        Root level archive init
        :return:
        """

    async def _dump_container_size(
            self, writer, container_len, container_type, params=None
    ):
        """
        Dumps container size - per element streaming
        :param writer:
        :param container_len:
        :param container_type:
        :param params:
        :return:
        """
        if not container_type or not container_type.FIX_SIZE:
            await dump_uvarint(writer, container_len)
        elif container_len != container_type.SIZE:
            raise ValueError(
                "Fixed size container has not defined size: %s" % container_type.SIZE
            )

    async def _dump_container_val(self, writer, elem, container_type, params=None):
        """
        Single elem dump
        :param writer:
        :param elem:
        :param container_type:
        :param params:
        :return:
        """
        elem_type = container_elem_type(container_type, params)
        await self.dump_field(writer, elem, elem_type, params[1:] if params else None)

    async def _dump_container(self, writer, container, container_type, params=None):
        """
        Dumps container of elements to the writer.

        :param writer:
        :param container:
        :param container_type:
        :param params:
        :return:
        """
        await self._dump_container_size(writer, len(container), container_type)

        elem_type = container_elem_type(container_type, params)

        for idx, elem in enumerate(container):
            try:
                self.tracker.push_index(idx)
                await self.dump_field(
                    writer, elem, elem_type, params[1:] if params else None
                )
                self.tracker.pop()
            except Exception as e:
                raise helpers.ArchiveException(e, tracker=self.tracker) from e

    async def _load_container(
            self, reader, container_type, params=None, container=None
    ):
        """
        Loads container of elements from the reader. Supports the container ref.
        Returns loaded container.

        :param reader:
        :param container_type:
        :param params:
        :param container:
        :return:
        """

        c_len = (
            container_type.SIZE
            if container_type.FIX_SIZE
            else await load_uvarint(reader)
        )
        if container and get_elem(container) and c_len != len(container):
            raise ValueError("Size mismatch")

        elem_type = container_elem_type(container_type, params)
        res = container if container else []
        for i in range(c_len):
            try:
                self.tracker.push_index(i)
                fvalue = await self.load_field(
                    reader,
                    elem_type,
                    params[1:] if params else None,
                    eref(res, i) if container else None,
                )
                self.tracker.pop()
            except Exception as e:
                raise helpers.ArchiveException(e, tracker=self.tracker) from e

            if not container:
                res.append(fvalue)
        return res

    async def _dump_tuple(self, writer, elem, elem_type, params=None):
        """
        Dumps tuple of elements to the writer.

        :param writer:
        :param elem:
        :param elem_type:
        :param params:
        :return:
        """
        if len(elem) != len(elem_type.f_specs()):
            raise ValueError(
                "Fixed size tuple has not defined size: %s" % len(elem_type.f_specs())
            )
        await dump_uvarint(writer, len(elem))

        elem_fields = params[0] if params else None
        if elem_fields is None:
            elem_fields = elem_type.f_specs()
        for idx, elem in enumerate(elem):
            try:
                self.tracker.push_index(idx)
                await self.dump_field(
                    writer, elem, elem_fields[idx], params[1:] if params else None
                )
                self.tracker.pop()
            except Exception as e:
                raise helpers.ArchiveException(e, tracker=self.tracker) from e

    async def _load_tuple(self, reader, elem_type, params=None, elem=None):
        """
        Loads tuple of elements from the reader. Supports the tuple ref.
        Returns loaded tuple.

        :param reader:
        :param elem_type:
        :param params:
        :param container:
        :return:
        """

        c_len = await load_uvarint(reader)
        if elem and c_len != len(elem):
            raise ValueError("Size mismatch")
        if c_len != len(elem_type.f_specs()):
            raise ValueError("Tuple size mismatch")

        elem_fields = params[0] if params else None
        if elem_fields is None:
            elem_fields = elem_type.f_specs()

        res = elem if elem else []
        for i in range(c_len):
            try:
                self.tracker.push_index(i)
                fvalue = await self.load_field(
                    reader,
                    elem_fields[i],
                    params[1:] if params else None,
                    eref(res, i) if elem else None,
                )
                self.tracker.pop()
            except Exception as e:
                raise helpers.ArchiveException(e, tracker=self.tracker) from e

            if not elem:
                res.append(fvalue)
        return res

    async def _dump_message_field(self, writer, msg, field, fvalue=None):
        """
        Dumps a message field to the writer. Field is defined by the message field specification.

        :param writer:
        :param msg:
        :param field:
        :param fvalue:
        :return:
        """
        fname, ftype, params = field[0], field[1], field[2:]
        fvalue = getattr(msg, fname, None) if fvalue is None else fvalue
        await self.dump_field(writer, fvalue, ftype, params)

    async def _load_message_field(self, reader, msg, field):
        """
        Loads message field from the reader. Field is defined by the message field specification.
        Returns loaded value, supports field reference.

        :param reader:
        :param msg:
        :param field:
        :return:
        """
        fname, ftype, params = field[0], field[1], field[2:]
        await self.load_field(reader, ftype, params, eref(msg, fname))

    async def _dump_variant(self, writer, elem, elem_type=None, params=None):
        """
        Dumps variant type to the writer.
        Supports both wrapped and raw variant.

        :param writer:
        :param elem:
        :param elem_type:
        :param params:
        :return:
        """
        if isinstance(elem, VariantType) or elem_type.WRAPS_VALUE:
            await dump_uint(writer, elem.variant_elem_type.VARIANT_CODE, 1)
            await self.dump_field(
                writer, getattr(elem, elem.variant_elem), elem.variant_elem_type
            )

        else:
            fdef = find_variant_fdef(elem_type, elem)
            await dump_uint(writer, fdef[1].VARIANT_CODE, 1)
            await self.dump_field(writer, elem, fdef[1])

    async def _load_variant(
            self, reader, elem_type, params=None, elem=None, wrapped=None
    ):
        """
        Loads variant type from the reader.
        Supports both wrapped and raw variant.

        :param reader:
        :param elem_type:
        :param params:
        :param elem:
        :param wrapped:
        :return:
        """
        is_wrapped = (
            (isinstance(elem, VariantType) or elem_type.WRAPS_VALUE)
            if wrapped is None
            else wrapped
        )
        if is_wrapped:
            elem = elem_type() if elem is None else elem

        tag = await load_uint(reader, 1)
        for field in elem_type.f_specs():
            ftype = field[1]
            if ftype.VARIANT_CODE == tag:
                fvalue = await self.load_field(
                    reader, ftype, field[2:], elem if not is_wrapped else None
                )
                if is_wrapped:
                    elem.set_variant(field[0], fvalue)
                return elem if is_wrapped else fvalue
        raise ValueError("Unknown tag: %s" % tag)


async def dump_blob(writer, elem, elem_type, params=None):
    """
    Dumps blob message to the writer.
    Supports both blob and raw value.

    :param writer:
    :param elem:
    :param elem_type:
    :param params:
    :return:
    """
    elem_is_blob = isinstance(elem, BlobType)
    elem_params = elem if elem_is_blob or elem_type is None else elem_type
    data = bytes(getattr(elem, BlobType.DATA_ATTR) if elem_is_blob else elem)

    if not elem_params.FIX_SIZE:
        await dump_uvarint(writer, len(elem))
    elif len(data) != elem_params.SIZE:
        raise ValueError('Fixed size blob has not defined size: %s' % elem_params.SIZE)
    await writer.awrite(data)


async def load_blob(reader, elem_type, params=None, elem=None):
    """
    Loads blob from reader to the element. Returns the loaded blob.

    :param reader:
    :param elem_type:
    :param params:
    :param elem:
    :return:
    """
    ivalue = elem_type.SIZE if elem_type.FIX_SIZE else await load_uvarint(reader)
    fvalue = bytearray(ivalue)
    await reader.areadinto(fvalue)

    if elem is None:
        return fvalue  # array by default

    elif isinstance(elem, BlobType):
        setattr(elem, elem_type.DATA_ATTR, fvalue)
        return elem

    else:
        elem.extend(fvalue)

    return elem


async def dump_unicode(writer, elem):
    """
    Dumps string as UTF8 encoded string
    :param writer:
    :param elem:
    :return:
    """
    await dump_uvarint(writer, len(elem))
    await writer.awrite(bytes(elem, 'utf8'))


async def load_unicode(reader):
    """
    Loads UTF8 string
    :param reader:
    :return:
    """
    ivalue = await load_uvarint(reader)
    fvalue = bytearray(ivalue)
    await reader.areadinto(fvalue)
    return str(fvalue, 'utf8')


def find_variant_fdef(elem_type, elem):
    fields = elem_type.f_specs()
    for x in fields:
        if isinstance(elem, x[1]):
            return x

    # Not direct hierarchy
    name = elem.__class__.__name__
    for x in fields:
        if name == x[1].__name__:
            return x

    raise ValueError("Unrecognized variant: %s" % elem)



from apps.monero.xmr.serialize.base_types import BoolType, SizeT, UInt64, UVarintType
from apps.monero.xmr.serialize.message_types import (
    ContainerType,
    MessageType,
    TupleType,
)
from apps.monero.xmr.serialize_messages.base import ECKey, ECPublicKey
from apps.monero.xmr.serialize_messages.ct_keys import CtKey


class MultisigKLRki(MessageType):
    @staticmethod
    def f_specs():
        return (("K", ECKey), ("L", ECKey), ("R", ECKey), ("ki", ECKey))


class OutputEntry(TupleType):
    @staticmethod
    def f_specs():
        return (UVarintType, CtKey)  # original: x.UInt64


class TxSourceEntry(MessageType):
    @staticmethod
    def f_specs():
        return (
            ("outputs", ContainerType, OutputEntry),
            ("real_output", SizeT),
            ("real_out_tx_key", ECPublicKey),
            ("real_out_additional_tx_keys", ContainerType, ECPublicKey),
            ("real_output_in_tx_index", UInt64),
            ("amount", UInt64),
            ("rct", BoolType),
            ("mask", ECKey),
            ("multisig_kLRki", MultisigKLRki),
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
Monero JSON codec.

Note: This is not optimized for memory usage.
Used for debugging.
'''

import base64
import collections
import json

from . import xmrserialize as x


def unescape_json_str(st):
    """
    Unescape Monero json encoded string
    /monero/external/rapidjson/reader.h

    :param st:
    :return:
    """
    c = 0
    ln = len(st)
    escape_chmap = {
        b'b': b'\b',
        b'f': b'\f',
        b'n': b'\n',
        b'r': b'\r',
        b't': b'\t',
        b'\\': b'\\',
        b'"': b'\"',
        b'/': b'\/'
    }

    ret = []

    def at(i):
        return st[i:i+1]

    while c < ln:
        if at(c) == b'\\':
            if at(c+1) == b'u':
                ret.append(bytes([int(st[c+2:c+6], 16)]))
                # ret.append(st[c:c+6].decode('unicode_escape').encode('utf8'))
                c += 6

            else:
                ret.append(escape_chmap[at(c+1)])
                c += 2

        else:
            ret.append(at(c))
            c += 1

    df = (b''.join(ret))
    return df


def escape_string_json(st):
    """
    Escaping string for json
    /Users/dusanklinec/workspace/monero/external/rapidjson/writer.h
    :param st:
    :return:
    """
    ret = []
    c = 0
    ln = len(st)

    escape_chmap = {
        b'\b': b'b',
        b'\f': b'f',
        b'\n': b'n',
        b'\r': b'r',
        b'\t': b't',
        b'\\': b'\\',
        b'\"': b'"',
        b'\/': b'/'
    }

    while c < ln:
        ch = st[c:c+1]
        if ch in escape_chmap:
            ret.append(b'\\' + escape_chmap[ch])
        elif ord(ch) < 0x20:
            ret.append(b'\\u%04X' % ord(ch))
        else:
            ret.append(ch)
        c += 1
    return b''.join(ret)


class Archive(x.Archive):

    def __init__(self, iobj, writing=True, **kwargs):
        super().__init__(iobj, writing, **kwargs)


class AutoJSONEncoder(json.JSONEncoder):
    """
    JSON encoder trying to_json() first
    """
    def default(self, obj):
        try:
            return obj.to_json()
        except AttributeError:
            return self.default_classic(obj)

    def default_classic(self, o):
        if isinstance(o, set):
            return list(o)
        elif isinstance(o, bytes):
            return str(o, 'utf8')
        else:
            return super(AutoJSONEncoder, self).default(o)


def json_dumps(obj, **kwargs):
    """
    Uses auto encoder to serialize the object
    :param obj:
    :param kwargs:
    :return:
    """
    return json.dumps(obj, cls=AutoJSONEncoder, **kwargs)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
XMR RPC serialization
WARNING: Not finished yet

{json, binary} <-> portable_storage <-> object

The portable storage is pretty much generic dict in python.
The original codec is not streaming in a sense directly translating input to object or vice versa.

Portable storage is an intermediate object.

Mainly related to:

/contrib/epee/include/serialization/keyvalue_serialization.h
/contrib/epee/include/serialization/keyvalue_serialization_overloads.h
/contrib/epee/include/storages/portable_storage.h
/contrib/epee/include/storages/portable_storage_base.h
/contrib/epee/include/storages/portable_storage_from_bin.h
/contrib/epee/include/storages/portable_storage_from_json.h
/contrib/epee/include/storages/portable_storage_to_bin.h
/contrib/epee/include/storages/portable_storage_to_json.h

WARNING: Not finished yet
'''

import binascii
import collections

from . import xmrserialize as x
from . import helpers
from .xmrserialize import eref


_UINT_BUFFER = bytearray(1)


class PortableStorageConsts:
    SIGNATUREA = 0x01011101
    SIGNATUREB = 0x01020101
    FORMAT_VER = 1


class PortableRawSizeMark:
    MASK = 0x03
    BYTE = 0
    WORD = 1
    DWORD = 2
    INT64 = 3


class SerializeType:
    INT64 = 1
    INT32 = 2
    INT16 = 3
    INT8 = 4
    UINT64 = 5
    UINT32 = 6
    UINT16 = 7
    UINT8 = 8
    DUOBLE = 9
    STRING = 10
    BOOL = 11
    OBJECT = 12
    ARRAY = 13
    ARRAY_FLAG = 0x80


SerializeTypeSize = {
    SerializeType.INT64: 8,
    SerializeType.INT32: 2,
    SerializeType.INT16: 2,
    SerializeType.INT8: 1,
    SerializeType.UINT64: 8,
    SerializeType.UINT32: 2,
    SerializeType.UINT16: 2,
    SerializeType.UINT8: 1,
    SerializeType.BOOL: 1,
}


XmrTypeMap = {
    x.UVarintType: SerializeType.UINT64,
    x.Int64: SerializeType.INT64,
    x.Int32: SerializeType.INT32,
    x.Int16: SerializeType.INT16,
    x.Int8: SerializeType.INT8,
    x.UInt64: SerializeType.UINT64,
    x.UInt32: SerializeType.UINT32,
    x.UInt16: SerializeType.UINT16,
    x.UInt8: SerializeType.UINT8,
    x.BoolType: SerializeType.BOOL,
    x.UnicodeType: SerializeType.STRING,
    x.ContainerType: SerializeType.ARRAY,
    x.MessageType: SerializeType.OBJECT,
    x.BlobType: SerializeType.STRING,
}


class BlobFieldWrapper(object):
    """Given field should be serialized/unserialized as blob"""
    def __init__(self, ftype, *args, **kwargs):
        self.ftype = ftype


class DefaultFieldWrapper(object):
    """Returns default if not set"""
    def __init__(self, ftype, default=None, *args, **kwargs):
        self.ftype = ftype
        self.default = default


class IModel(object):
    def __init__(self, val, ser_type=None):
        self.val = val
        self.type = ser_type

    def __repr__(self):
        return '%r' % self.val

    def __eq__(self, other):
        if isinstance(other, IModel):
            return self.type == other.type and self.val == other.val
        else:
            return self.val == other

    def to_json(self):
        return self.val

    def val(self):
        return self.val()


class ArrayModel(IModel):
    def __init__(self, val, ser_type=None):
        super().__init__(val, ser_type)
        self.current_idx = 0

    def __str__(self):
        return 'Arr[%s; %s]' % (self.type, self.val)

    def __len__(self):
        return len(self.val)

    def __getitem__(self, item):
        return self.val[item]

    def __setitem__(self, key, value):
        self.val[key] = value

    def __iter__(self):
        self.current_idx = 0
        return self

    def __next__(self):
        if self.current_idx >= len(self):
            raise StopIteration
        else:
            self.current_idx += 1
            return self[self.current_idx - 1]


class IntegerModel(IModel):
    def __str__(self):
        return 'Int[%s; %s]' % (self.type, self.val)


class BlobModel(IModel):
    def __str__(self):
        return 'Blob[%s; %s]' % (self.type, self.val)


class NoSetSentinel(object):
    pass


def int_mark_to_size(int_type):
    if int_type == PortableRawSizeMark.BYTE:
        return 1
    elif int_type == PortableRawSizeMark.WORD:
        return 2
    elif int_type == PortableRawSizeMark.DWORD:
        return 4
    elif int_type == PortableRawSizeMark.INT64:
        return 8
    else:
        raise ValueError('Unknown int type')


def type_to_size(obj_type):
    return SerializeTypeSize[obj_type]


def xmr_type_to_type(elem_type):
    if isinstance(elem_type, int):
        return elem_type
    elif elem_type == object or elem_type == type:
        raise ValueError('Canot convert Xmr type: %s' % elem_type)
    elif elem_type in XmrTypeMap:
        return XmrTypeMap[elem_type]
    elif isinstance(elem_type, x.XmrType) and elem_type.__class__ in XmrTypeMap:
        return XmrTypeMap[elem_type.__class__]
    elif len(elem_type.__bases__) > 0:
        return xmr_type_to_type(elem_type.__bases__[0])
    else:
        raise ValueError('Cannot convert Xmr type %s' % elem_type)


async def dump_varint_t(writer, type_or, pv):
    """
    Binary dump of the integer of given type

    :param writer:
    :param type_or:
    :param pv:
    :return:
    """
    width = int_mark_to_size(type_or)
    n = (pv << 2) | type_or

    buffer = _UINT_BUFFER
    for _ in range(width):
        buffer[0] = n & 0xff
        await writer.awrite(buffer)
        n >>= 8

    return width


async def dump_varint(writer, val):
    """
    Binary dump of the variable size integer

    :param writer:
    :param val:
    :return:
    """
    if val <= 63:
        return await dump_varint_t(writer, PortableRawSizeMark.BYTE, val)
    elif val <= 16383:
        return await dump_varint_t(writer, PortableRawSizeMark.WORD, val)
    elif val <= 1073741823:
        return await dump_varint_t(writer, PortableRawSizeMark.DWORD, val)
    else:
        if val > 4611686018427387903:
            raise ValueError('Int too big')
        return await dump_varint_t(writer, PortableRawSizeMark.INT64, val)


async def load_varint(reader):
    """
    Binary load of variable size integer serialized by dump_varint

    :param reader:
    :return:
    """
    buffer = _UINT_BUFFER

    await reader.areadinto(buffer)
    width = int_mark_to_size(buffer[0] & PortableRawSizeMark.MASK)
    result = buffer[0]

    shift = 8
    for _ in range(width-1):
        await reader.areadinto(buffer)
        result += buffer[0] << shift
        shift += 8
    return result >> 2


async def dump_string(writer, val):
    """
    Binary string dump

    :param writer:
    :param val:
    :return:
    """
    await dump_varint(writer, len(val))
    await writer.awrite(val)


async def load_string(reader):
    """
    Loads string from binary stream

    :param reader:
    :return:
    """
    ivalue = await load_varint(reader)
    fvalue = bytearray(ivalue)
    await reader.areadinto(fvalue)
    return bytes(fvalue)


async def dump_blob(writer, elem, elem_type, params=None):
    """
    Dumps blob to a binary stream

    :param writer:
    :param elem:
    :param elem_type:
    :param params:
    :return:
    """
    elem_is_blob = isinstance(elem, x.BlobType)
    data = bytes(getattr(elem, x.BlobType.DATA_ATTR) if elem_is_blob else elem)
    await dump_varint(writer, len(elem))
    await writer.awrite(data)


async def load_blob(reader, elem_type, params=None, elem=None):
    """
    Loads blob from reader to the element. Returns the loaded blob.

    :param reader:
    :param elem_type:
    :param params:
    :param elem:
    :return:
    """
    ivalue = await load_varint(reader)
    fvalue = bytearray(ivalue)
    await reader.areadinto(fvalue)

    if elem is None:
        return fvalue  # array by default

    elif isinstance(elem, x.BlobType):
        setattr(elem, elem_type.DATA_ATTR, fvalue)
        return elem

    else:
        elem.extend(fvalue)

    return elem


#
# Archive
#


class Archive(x.Archive):
    def __init__(self, iobj, writing=True, modeled=True, **kwargs):
        super().__init__(iobj, writing, **kwargs)
        self.modeled = modeled

    async def root(self):
        """
        Root level init
        :return:
        """
        if self.writing:
            await x.dump_uint(self.iobj, PortableStorageConsts.SIGNATUREA, 4)
            await x.dump_uint(self.iobj, PortableStorageConsts.SIGNATUREB, 4)
            await x.dump_uint(self.iobj, PortableStorageConsts.FORMAT_VER, 1)

        else:
            sig_a = await x.load_uint(self.iobj, 4)
            sig_b = await x.load_uint(self.iobj, 4)
            ver = await x.load_uint(self.iobj, 1)

            if sig_a != PortableStorageConsts.SIGNATUREA:
                raise ValueError('Signature A error')
            if sig_b != PortableStorageConsts.SIGNATUREB:
                raise ValueError('Signature A error')
            if ver != PortableStorageConsts.FORMAT_VER:
                raise ValueError('Unsupported version')

    async def section(self, sec=None):
        """
        Section / dict serialization
        :return:
        """
        if self.writing:
            await dump_varint(self.iobj, len(sec))
            for key in sec:
                await self.section_name(key)
                await self.storage_entry(sec[key])

        else:
            sec = {} if sec is None else sec

            count = await load_varint(self.iobj)
            for idx in range(count):
                sec_name = await self.section_name()
                val = await self.storage_entry()
                sec[sec_name] = val

            return sec

    async def section_name(self, sec_name=None):
        """
        Section name
        :param sec_name:
        :return:
        """
        if self.writing:
            fvalue = sec_name.encode('ascii')
            await x.dump_uint(self.iobj, len(fvalue), 1)
            await self.iobj.awrite(bytearray(fvalue))

        else:
            ivalue = await x.load_uint(self.iobj, 1)
            fvalue = bytearray(ivalue)
            await self.iobj.areadinto(fvalue)
            return bytes(fvalue).decode('ascii')

    async def storage_entry(self, entry=None, ent_type=None):
        if self.writing:
            oentry = entry
            if ent_type is None:
                ent_type, entry = self.det_entry_model(entry)

            if ent_type & SerializeType.ARRAY_FLAG:
                return await self.entry(ent_type=ent_type, elem=oentry)

            else:
                await x.dump_uint(self.iobj, ent_type, 1)
                return await self.entry(ent_type=ent_type, elem=entry)

        else:
            ent_type = await x.load_uint(self.iobj, 1)
            return await self.entry(ent_type)

    async def array(self, container=None, container_type=None, params=None):
        if self.writing:
            if container_type is None and not isinstance(container, ArrayModel):
                raise ValueError('Unknown container type serialization')
            if container_type is None:
                container_type = container.type
                container = container.val

            entry_type = container_type & (~ SerializeType.ARRAY_FLAG)
            container_type |= SerializeType.ARRAY_FLAG

            await x.dump_uint(self.iobj, container_type, 1)
            await dump_varint(self.iobj, len(container))
            for i in container:
                await self.entry(entry_type, elem=i)

        else:
            if container_type is None:
                container_type = await x.load_uint(self.iobj, 1)

            container_type &= ~SerializeType.ARRAY_FLAG

            c_len = await load_varint(self.iobj)
            res = container if container else []
            for i in range(c_len):
                fval = await self.entry(container_type, x.eref(res, i) if container else None)
                if not container:
                    res.append(fval)

            return res if not self.modeled else ArrayModel(res, container_type)

    async def entry(self, ent_type, elem=None):
        if self.writing:
            oelem = elem
            elem = elem if not isinstance(elem, IModel) else elem.val

            if ent_type in SerializeTypeSize:
                return await x.dump_uint(self.iobj, elem, type_to_size(ent_type))
            elif ent_type == SerializeType.DUOBLE:
                raise ValueError('Not supported')
            elif ent_type == SerializeType.STRING:
                return await self.unicode_type(elem)
            elif ent_type == SerializeType.OBJECT:
                return await self.section(elem)
            elif ent_type == SerializeType.ARRAY:
                return await self.array(oelem)
            elif ent_type & SerializeType.ARRAY_FLAG:
                return await self.array(elem, container_type=ent_type & (~SerializeType.ARRAY_FLAG))
            else:
                raise ValueError('Unrecognized type 0x%x' % ent_type)

        else:
            if ent_type in SerializeTypeSize:
                fval = await x.load_uint(self.iobj, type_to_size(ent_type))
                return fval if not self.modeled else IntegerModel(fval, ent_type)
            elif ent_type == SerializeType.DUOBLE:
                raise ValueError('Not supported')
            elif ent_type == SerializeType.STRING:
                return await self.unicode_type()
            elif ent_type == SerializeType.OBJECT:
                return await self.section()
            elif ent_type == SerializeType.ARRAY:
                return await self.array()
            elif ent_type & SerializeType.ARRAY_FLAG:
                return await self.array(container_type=ent_type & (~SerializeType.ARRAY_FLAG))
            else:
                raise ValueError('Unrecognized type 0x%x' % ent_type)

    async def unicode_type(self, elem=None):
        if self.writing:
            return await dump_string(self.iobj, elem)

        else:
            return await load_string(self.iobj)

    @staticmethod
    def det_entry_model(entry):
        if isinstance(entry, IntegerModel):
            return entry.type, entry.val
        elif isinstance(entry, ArrayModel):
            return entry.type | SerializeType.ARRAY_FLAG, entry.val
        elif isinstance(entry, IModel):
            return entry.type, entry.val
        elif isinstance(entry, bytes):
            return SerializeType.STRING, entry
        elif isinstance(entry, bytearray):
            return SerializeType.STRING, bytes(entry)
        elif isinstance(entry, (dict, x.MessageType)):
            return SerializeType.OBJECT, entry
        elif isinstance(entry, (list, tuple)):
            return SerializeType.OBJECT | SerializeType.ARRAY_FLAG, entry  # fallback to obj
        else:
            raise ValueError('Unknown: %r' % entry)


#
# Blob serializer
#

class Blobber(object):
    """
    Serializing structures to Blob for KV_SERIALIZE.
    """
    def __init__(self, iobj=None, writing=True, data=None, **kwargs):
        self.writing = writing
        self.tracker = helpers.Tracker()
        self.iobj = x.MemoryReaderWriter() if iobj is not None else iobj
        if data is not None:
            self.iobj = x.MemoryReaderWriter(bytearray(data))

    async def uvarint(self, elem):
        """
        Uvarint untouched
        :param elem:
        :return:
        """
        if self.writing:
            return await dump_varint(self.iobj, elem)
        else:
            return await load_varint(self.iobj)

    async def uint(self, elem, elem_type, params=None):
        """
        Integer types
        :param elem:
        :param elem_type:
        :param params:
        :return:
        """
        if self.writing:
            return await x.dump_uint(self.iobj, elem, elem_type.WIDTH)
        else:
            return await x.load_uint(self.iobj, elem_type.WIDTH)

    async def unicode_type(self, elem):
        """
        Unicode type
        :param elem:
        :return:
        """
        if self.writing:
            return await x.dump_unicode(self.iobj, elem)
        else:
            return await x.load_unicode(self.iobj)

    async def blob(self, elem=None, elem_type=None, params=None):
        """
        Loads/dumps blob
        :return:
        """
        elem_type = elem_type if elem_type else elem.__class__
        if hasattr(elem_type, 'blob_serialize'):
            elem = elem_type() if elem is None else elem
            return await elem.blob_serialize(self, elem=elem, elem_type=elem_type, params=params)

        if self.writing:
            return await x.dump_blob(self.iobj, elem=elem, elem_type=elem_type, params=params)
        else:
            return await x.load_blob(self.iobj, elem_type=elem_type, params=params, elem=elem)

    async def container(self, container=None, container_type=None, params=None):
        """
        Loads/dumps container
        :return:
        """
        if hasattr(container_type, 'blob_serialize'):
            container = container_type() if container is None else container
            return await container.blob_serialize(self, elem=container, elem_type=container_type, params=params)

        # Container entry version + container
        if self.writing:
            return await self.container_dump(container, container_type, params)
        else:
            return await self.container_load(container_type, params=params, container=container)

    async def get_element_size(self, elem_type, elem=None, params=None):
        """
        Determines element size after blob serialization.
        Can be quite tricky for complicated structures - another object required to implement sizeof()

        :param elem_type:
        :param elem:
        :param params:
        :return:
        """
        if hasattr(elem_type, 'blob_size'):
            elem = elem_type() if elem is None else elem
            return await elem.blob_size(self)

        if issubclass(elem_type, x.UVarintType):
            raise helpers.ArchiveException('Unknown size for varint')

        elif issubclass(elem_type, x.IntType):
            return elem_type.WIDTH

        elif issubclass(elem_type, x.BlobType):
            if elem is not None:
                return len(elem)
            if not elem_type.FIX_SIZE:
                raise helpers.ArchiveException('Unknown size for blob')
            return elem_type.SIZE

        elif issubclass(elem_type, x.UnicodeType):
            raise helpers.ArchiveException('Unknown size for string')

        elif issubclass(elem_type, x.VariantType):
            raise helpers.ArchiveException('Unknown size for variant')

        elif issubclass(elem_type, x.ContainerType):  # container ~ simple list
            celem_type = x.container_elem_type(elem_type, params)
            if elem is not None:
                return len(elem) * await self.get_element_size(elem_type=celem_type, elem=elem[0] if len(elem) > 0 else None)
            if not elem_type.FIX_SIZE:
                raise helpers.ArchiveException('Unknown size for container')
            return elem_type.SIZE * await self.get_element_size(elem_type=celem_type)

        elif issubclass(elem_type, x.TupleType):  # tuple ~ simple list
            acc = 0
            for t in elem_type.f_specs():
                acc += await self.get_element_size(t)
            return acc

        elif issubclass(elem_type, x.MessageType):
            acc = 0
            for t in elem_type.f_specs():
                acc += await self.get_element_size(t[1], params=t[2:])
            return acc

        else:
            raise TypeError

    async def container_dump(self, container, container_type, params=None):
        """
        Dumps container of elements to the writer.
        Blob array writer as in XMRRPC is serialized without size serialization

        :param container:
        :param container_type:
        :param params:
        :param obj:
        :return:
        """
        elem_type = x.container_elem_type(container_type, params)
        if container is None:
            return

        for idx, elem in enumerate(container):
            try:
                self.tracker.push_index(idx)
                await self._dump_field(elem, elem_type, params[1:] if params else None)
                self.tracker.pop()
            except Exception as e:
                raise helpers.ArchiveException(e, tracker=self.tracker) from e

    async def container_load(self, container_type, params=None, container=None, obj=None):
        """
        Loads container of elements from the reader. Supports the container ref.
        Returns loaded container.
        Blob array writer as in XMRRPC is serialized without size serialization.

        :param container_type:
        :param params:
        :param container:
        :param obj:
        :return:
        """
        elem_type = x.container_elem_type(container_type, params)
        elem_size = await self.get_element_size(elem_type=elem_type, params=params)

        # If container is of fixed size we know the size to load from the input.
        # Otherwise we have to read to the end
        data_left = len(self.iobj.buffer)
        c_len = container_type.SIZE
        if not container_type.FIX_SIZE:
            if data_left == 0:
                return None
            if data_left % elem_size != 0:
                raise helpers.ArchiveException('Container size mod elem size not 0')
            c_len = data_left // elem_size

        res = container if container else []
        for i in range(c_len):
            try:
                self.tracker.push_index(i)
                fvalue = await self._load_field(elem_type,
                                                params[1:] if params else None,
                                                x.eref(res, i) if container else None)
                self.tracker.pop()
            except Exception as e:
                raise helpers.ArchiveException(e, tracker=self.tracker) from e

            if not container:
                res.append(fvalue)
        return res

    async def tuple(self, elem=None, elem_type=None, params=None):
        """
        Loads/dumps tuple
        :return:
        """
        if hasattr(elem_type, 'blob_serialize'):
            container = elem_type() if elem is None else elem
            return await container.blob_serialize(self, elem=elem, elem_type=elem_type, params=params)

        if self.writing:
            return await self.dump_tuple(elem, elem_type, params)
        else:
            return await self.load_tuple(elem_type, params=params, elem=elem)

    async def dump_tuple(self, elem, elem_type, params=None, obj=None):
        """
        Dumps tuple of elements to the writer.

        :param elem:
        :param elem_type:
        :param params:
        :param obj:
        :return:
        """
        raise helpers.ArchiveException('Not yet implemented')

    async def load_tuple(self, elem_type, params=None, elem=None, obj=None):
        """
        Loads tuple of elements from the reader. Supports the tuple ref.
        Returns loaded tuple.

        :param elem_type:
        :param params:
        :param elem:
        :param obj:
        :return:
        """
        raise helpers.ArchiveException('Not yet implemented')

    async def variant(self, elem=None, elem_type=None, params=None):
        """
        Loads/dumps variant type
        :param elem:
        :param elem_type:
        :param params:
        :param obj:
        :return:
        """
        elem_type = elem_type if elem_type else elem.__class__

        if hasattr(elem_type, 'blob_serialize'):
            elem = elem_type() if elem is None else elem
            return await elem.bob_serialize(self, elem=elem, elem_type=elem_type, params=params)

        if self.writing:
            return await self.dump_variant(elem=elem,
                                           elem_type=elem_type if elem_type else elem.__class__, params=params)
        else:
            return await self.load_variant(elem_type=elem_type if elem_type else elem.__class__,
                                           params=params, elem=elem)

    async def dump_variant(self, elem, elem_type=None, params=None, obj=None):
        """
        Dumps variant type to the writer.
        Supports both wrapped and raw variant.

        :param elem:
        :param elem_type:
        :param params:
        :param obj:
        :return:
        """
        raise helpers.ArchiveException('Not yet implemented')

    async def load_variant(self, elem_type, params=None, elem=None, wrapped=None, obj=None):
        """
        Loads variant type from the reader.
        Supports both wrapped and raw variant.

        :param elem_type:
        :param params:
        :param elem:
        :param wrapped:
        :param obj:
        :return:
        """
        raise helpers.ArchiveException('Not yet implemented')

    async def message(self, msg, msg_type=None):
        """
        Loads/dumps message
        :param msg:
        :param msg_type:
        :param obj:
        :return:
        """
        elem_type = msg_type if msg_type is not None else msg.__class__

        if hasattr(elem_type, 'blob_serialize'):
            msg = elem_type() if msg is None else msg
            return await msg.blob_serialize(self)

        fields = elem_type.f_specs()
        for field in fields:
            await self.message_field(msg=msg, field=field)

        return msg

    async def message_field(self, msg, field, fvalue=None, obj=None):
        """
        Dumps/Loads message field
        :param msg:
        :param field:
        :param fvalue: explicit value for dump
        :param obj:
        :return:
        """
        fname, ftype, params = field[0], field[1], field[2:]
        try:
            self.tracker.push_field(fname)
            if self.writing:
                fvalue = getattr(msg, fname, None) if fvalue is None else fvalue
                await self._dump_field(fvalue, ftype, params)

            else:
                await self._load_field(ftype, params, x.eref(msg, fname))

            self.tracker.pop()

        except Exception as e:
            raise helpers.ArchiveException(e, tracker=self.tracker) from e

    async def message_fields(self, msg, fields, obj=None):
        """
        Load/dump individual message fields
        :param msg:
        :param fields:
        :param obj:
        :return:
        """
        for field in fields:
            await self.message_field(msg, field, obj)
        return msg

    async def field(self, elem=None, elem_type=None, params=None):
        """
        Archive field
        :param elem:
        :param elem_type:
        :param params:
        :return:
        """
        elem_type = elem_type if elem_type else elem.__class__
        fvalue = None

        src = elem
        if issubclass(elem_type, x.UVarintType):
            fvalue = await self.uvarint(x.get_elem(src))

        elif issubclass(elem_type, x.IntType):
            fvalue = await self.uint(elem=x.get_elem(src), elem_type=elem_type, params=params)

        elif issubclass(elem_type, x.BlobType):
            fvalue = await self.blob(elem=x.get_elem(src), elem_type=elem_type, params=params)

        elif issubclass(elem_type, x.UnicodeType):
            fvalue = await self.unicode_type(x.get_elem(src))

        elif issubclass(elem_type, x.VariantType):
            fvalue = await self.variant(elem=x.get_elem(src), elem_type=elem_type, params=params)

        elif issubclass(elem_type, x.ContainerType):  # container ~ simple list
            fvalue = await self.container(container=x.get_elem(src), container_type=elem_type, params=params)

        elif issubclass(elem_type, x.TupleType):  # tuple ~ simple list
            fvalue = await self.tuple(elem=x.get_elem(src), elem_type=elem_type, params=params)

        elif issubclass(elem_type, x.MessageType):
            fvalue = await self.message(x.get_elem(src), msg_type=elem_type)

        else:
            raise TypeError

        return fvalue if self.writing else x.set_elem(elem, fvalue)

    async def _dump_field(self, elem, elem_type, params=None):
        return await self.field(elem=elem, elem_type=elem_type, params=params)

    async def _load_field(self, elem_type, params=None, elem=None):
        return await self.field(elem=elem, elem_type=elem_type, params=params)

    async def blobize(self, elem=None, elem_type=None, params=None):
        """
        Main blobbing
        :param elem:
        :param elem_type:
        :param params:
        :return:
        """
        if self.writing:
            await self.field(elem=elem, elem_type=elem_type, params=params)
            return bytes(self.iobj.buffer)
        else:
            return await self.field(elem=elem, elem_type=elem_type, params=params)


class Modeler(object):
    """
    Converting message classes to models and vice versa.
    Writing  = transforming message to model.
    !Writing = transforming model to message.
    """

    def __init__(self, writing=True, hexlify=False, modelize=True, strict_load=False, **kwargs):
        self.writing = writing
        self.hexlify = hexlify
        self.modelize = modelize
        self.strict_load = strict_load
        self.tracker = helpers.Tracker()

    @staticmethod
    def to_bytes(elem):
        if isinstance(elem, bytearray):
            return bytes(elem)
        if isinstance(elem, str):
            return elem.encode('utf8')
        if isinstance(elem, list):
            return bytes(elem)
        return elem

    async def uvarint(self, elem):
        """
        Uvarint untouched
        :param elem:
        :return:
        """
        return elem

    async def uint(self, elem, elem_type, params=None):
        """
        Integer types
        :param elem:
        :param elem_type:
        :param params:
        :return:
        """
        if self.writing:
            return IntegerModel(elem, elem_type.WIDTH) if self.modelize else elem
        else:
            return elem.val if isinstance(elem, IModel) else elem

    async def unicode_type(self, elem):
        """
        Unicode type
        :param elem:
        :return:
        """
        if self.writing:
            return Modeler.to_bytes(elem)

        else:
            return elem

    async def blob(self, elem=None, elem_type=None, params=None):
        """
        Loads/dumps blob
        :return:
        """
        elem_type = elem_type if elem_type else elem.__class__
        if hasattr(elem_type, 'kv_serialize'):
            elem = elem_type() if elem is None else elem
            return await elem.kv_serialize(self, elem=elem, elem_type=elem_type, params=params)

        if self.writing:
            elem_is_blob = isinstance(elem, x.BlobType)
            data = getattr(elem, x.BlobType.DATA_ATTR) if elem_is_blob else elem
            if data is None:
                return NoSetSentinel()
            if len(data) == 0:
                return b''

            fval = Modeler.to_bytes(data)
            if self.hexlify:
                return binascii.hexlify(fval).decode('ascii')
            else:
                return fval

        else:
            if elem is None:
                return NoSetSentinel()
            if self.hexlify:
                return bytes(binascii.unhexlify(elem))
            else:
                return bytes(elem)

    async def container(self, container=None, container_type=None, params=None, obj=None):
        """
        Loads/dumps container
        :return:
        """
        if hasattr(container_type, 'kv_serialize'):
            container = container_type() if container is None else container
            return await container.kv_serialize(self, elem=container, elem_type=container_type, params=params, obj=obj)

        # Container entry version + container
        if self.writing:
            return await self.container_dump(container, container_type, params, obj=obj)
        else:
            return await self.container_load(container_type, params=params, container=container, obj=obj)

    async def container_dump(self, container, container_type, params=None, obj=None):
        """
        Dumps container of elements to the writer.

        :param container:
        :param container_type:
        :param params:
        :param obj:
        :return:
        """
        elem_type = x.container_elem_type(container_type, params)
        obj = [] if not x.has_elem(obj) else x.get_elem(obj)

        # todo: pod container, just concat blobs / serialized content together. loading = size / elem size...
        if container is None:  # todo: reconsider
            return NoSetSentinel()  # if not self.modelize else ArrayModel(obj, xmr_type_to_type(elem_type))

        for idx, elem in enumerate(container):
            try:
                self.tracker.push_index(idx)
                fvalue = await self._dump_field(elem, elem_type, params[1:] if params else None)
                self.tracker.pop()

            except Exception as e:
                raise helpers.ArchiveException(e, tracker=self.tracker) from e

            if not isinstance(fvalue, NoSetSentinel):
                obj.append(fvalue)

        return obj if not self.modelize else ArrayModel(obj, xmr_type_to_type(elem_type))

    async def container_load(self, container_type, params=None, container=None, obj=None):
        """
        Loads container of elements from the reader. Supports the container ref.
        Returns loaded container.

        :param container_type:
        :param params:
        :param container:
        :param obj:
        :return:
        """
        if isinstance(obj, IModel):
            obj = obj.val

        if obj is None:
            return NoSetSentinel()

        c_len = len(obj)
        elem_type = params[0] if params else None
        if elem_type is None:
            elem_type = container_type.ELEM_TYPE

        res = container if container else []
        for i in range(c_len):
            try:
                self.tracker.push_index(i)
                fvalue = await self._load_field(elem_type,
                                                params[1:] if params else None,
                                                x.eref(res, i) if container else None, obj=obj[i])
                self.tracker.pop()

            except Exception as e:
                raise helpers.ArchiveException(e, tracker=self.tracker) from e

            if not container and not isinstance(fvalue, NoSetSentinel):
                res.append(fvalue)

        return res

    async def tuple(self, elem=None, elem_type=None, params=None, obj=None):
        """
        Loads/dumps tuple
        :return:
        """
        if hasattr(elem_type, 'kv_serialize'):
            container = elem_type() if elem is None else elem
            return await container.kv_serialize(self, elem=elem, elem_type=elem_type, params=params, obj=obj)

        # TODO: if modeled return as 0=>, 1=>, ...
        if self.writing:
            return await self.dump_tuple(elem, elem_type, params, obj=obj)
        else:
            return await self.load_tuple(elem_type, params=params, elem=elem, obj=obj)

    async def dump_tuple(self, elem, elem_type, params=None, obj=None):
        """
        Dumps tuple of elements to the writer.

        :param elem:
        :param elem_type:
        :param params:
        :param obj:
        :return:
        """
        if len(elem) != len(elem_type.f_specs()):
            raise ValueError('Fixed size tuple has not defined size: %s' % len(elem_type.f_specs()))

        elem_fields = params[0] if params else None
        if elem_fields is None:
            elem_fields = elem_type.f_specs()

        obj = [] if obj is None else x.get_elem(obj)
        for idx, elem in enumerate(elem):
            try:
                self.tracker.push_index(idx)
                fvalue = await self._dump_field(elem, elem_fields[idx], params[1:] if params else None)
                obj.append(fvalue)
                self.tracker.pop()

            except Exception as e:
                raise helpers.ArchiveException(e, tracker=self.tracker) from e

        return obj

    async def load_tuple(self, elem_type, params=None, elem=None, obj=None):
        """
        Loads tuple of elements from the reader. Supports the tuple ref.
        Returns loaded tuple.

        :param elem_type:
        :param params:
        :param elem:
        :param obj:
        :return:
        """
        if obj is None:
            return None

        elem_fields = params[0] if params else None
        if elem_fields is None:
            elem_fields = elem_type.f_specs()

        c_len = len(obj)
        if len(elem_fields) != c_len:
            raise ValueError('Size mismatch')

        res = elem if elem else []
        for i in range(len(elem_fields)):
            try:
                self.tracker.push_index(i)
                fvalue = await self._load_field(params[1:] if params else None,
                                                x.eref(res, i) if elem else None,
                                                obj=obj[i])
                self.tracker.pop()

                if not elem:
                    res.append(fvalue)

            except Exception as e:
                raise helpers.ArchiveException(e, tracker=self.tracker) from e

        return res

    async def variant(self, elem=None, elem_type=None, params=None, obj=None):
        """
        Loads/dumps variant type
        :param elem:
        :param elem_type:
        :param params:
        :param obj:
        :return:
        """
        elem_type = elem_type if elem_type else elem.__class__

        if hasattr(elem_type, 'kv_serialize'):
            elem = elem_type() if elem is None else elem
            return await elem.kv_serialize(self, elem=elem, elem_type=elem_type, params=params, obj=obj)

        if self.writing:
            return await self.dump_variant(elem=elem,
                                           elem_type=elem_type if elem_type else elem.__class__, params=params, obj=obj)
        else:
            return await self.load_variant(elem_type=elem_type if elem_type else elem.__class__,
                                           params=params, elem=elem, obj=obj)

    async def dump_variant(self, elem, elem_type=None, params=None, obj=None):
        """
        Dumps variant type to the writer.
        Supports both wrapped and raw variant.

        :param elem:
        :param elem_type:
        :param params:
        :param obj:
        :return:
        """
        fvalue = None
        if isinstance(elem, x.VariantType) or elem_type.WRAPS_VALUE:
            try:
                self.tracker.push_variant(elem.variant_elem_type)
                fvalue = {
                    elem.variant_elem: await self._dump_field(getattr(elem, elem.variant_elem), elem.variant_elem_type, obj=obj)
                }
                self.tracker.pop()

            except Exception as e:
                raise helpers.ArchiveException(e, tracker=self.tracker) from e

        else:
            try:
                fdef = elem_type.find_fdef(elem_type.f_specs(), elem)
                self.tracker.push_variant(fdef[1])
                fvalue = {
                    fdef[0]: await self._dump_field(elem, fdef[1], obj=obj)
                }
                self.tracker.pop()

            except Exception as e:
                raise helpers.ArchiveException(e, tracker=self.tracker) from e

        return fvalue

    async def load_variant(self, elem_type, params=None, elem=None, wrapped=None, obj=None):
        """
        Loads variant type from the reader.
        Supports both wrapped and raw variant.

        :param elem_type:
        :param params:
        :param elem:
        :param wrapped:
        :param obj:
        :return:
        """
        is_wrapped = elem_type.WRAPS_VALUE if wrapped is None else wrapped

        if is_wrapped:
            elem = elem_type() if elem is None else elem

        fname = list(obj.keys())[0]
        for field in elem_type.f_specs():
            if field[0] != fname:
                continue

            try:
                self.tracker.push_variant(field[1])
                fvalue = await self._load_field(field[1], field[2:], elem if not is_wrapped else None, obj=obj[fname])
                self.tracker.pop()

            except Exception as e:
                raise helpers.ArchiveException(e, tracker=self.tracker) from e

            if is_wrapped:
                elem.set_variant(field[0], fvalue)

            return elem if is_wrapped else fvalue
        raise ValueError('Unknown tag: %s' % fname)

    async def message(self, msg, msg_type=None, obj=None):
        """
        Loads/dumps message
        :param msg:
        :param msg_type:
        :param obj:
        :return:
        """
        elem_type = msg_type if msg_type is not None else msg.__class__
        obj = collections.OrderedDict() if not x.has_elem(obj) else x.get_elem(obj)

        if hasattr(elem_type, 'kv_serialize'):
            msg = elem_type() if msg is None else msg
            return await msg.kv_serialize(self, obj=obj)

        fields = elem_type.f_specs()
        for field in fields:
            await self.message_field(msg=msg, field=field, obj=obj)

        return obj if self.writing else msg

    async def message_field(self, msg, field, fvalue=None, obj=None):
        """
        Dumps/Loads message field
        :param msg:
        :param field:
        :param fvalue: explicit value for dump
        :param obj:
        :return:
        """
        fname, ftype, params = field[0], field[1], field[2:]

        try:
            self.tracker.push_field(field[0])

            if self.writing:
                if msg is None:
                    return None
                fvalue = getattr(msg, fname, None) if fvalue is None else fvalue
                await self._dump_field(fvalue, ftype, params, obj=eref(obj, fname, True))
            else:
                oval = obj[fname] if self.strict_load else (obj[fname] if fname in obj else None)
                await self._load_field(ftype, params, x.eref(msg, fname), obj=oval)

            self.tracker.pop()

        except Exception as e:
            raise helpers.ArchiveException(e, tracker=self.tracker) from e

    async def message_fields(self, msg, fields, obj=None):
        """
        Load/dump individual message fields
        :param msg:
        :param fields:
        :param obj:
        :return:
        """
        for field in fields:
            await self.message_field(msg, field, obj)
        return msg

    async def field(self, elem=None, elem_type=None, params=None, obj=None):
        """
        Archive field
        :param elem:
        :param elem_type:
        :param params:
        :param obj:
        :return:
        """
        elem_type = elem_type if elem_type else elem.__class__
        fvalue = None

        src = elem if self.writing else obj
        dst = obj if self.writing else elem

        # TODO: optional elem, default value for deserialization...

        # Blob wrapper. Underlying structure should be serialized as blob.
        if x.is_type(elem_type, BlobFieldWrapper):
            blobber = Blobber(self.writing, data=x.get_elem(src) if not self.writing else None)
            fvalue = await blobber.blobize(elem=x.get_elem(src), elem_type=elem_type.ftype, params=params)
            fvalue = NoSetSentinel() if fvalue is None or len(fvalue) == 0 else fvalue

        elif issubclass(elem_type, x.UVarintType):
            fvalue = await self.uvarint(x.get_elem(src))

        elif issubclass(elem_type, x.IntType):
            fvalue = await self.uint(elem=x.get_elem(src), elem_type=elem_type, params=params)

        elif issubclass(elem_type, x.BlobType):
            fvalue = await self.blob(elem=x.get_elem(src), elem_type=elem_type, params=params)

        elif issubclass(elem_type, x.UnicodeType):
            fvalue = await self.unicode_type(x.get_elem(src))

        elif issubclass(elem_type, x.VariantType):
            fvalue = await self.variant(elem=x.get_elem(src), elem_type=elem_type, params=params, obj=dst)

        elif issubclass(elem_type, x.ContainerType):  # container ~ simple list
            fvalue = await self.container(container=x.get_elem(src), container_type=elem_type, params=params, obj=dst)

        elif issubclass(elem_type, x.TupleType):  # tuple ~ simple list
            fvalue = await self.tuple(elem=x.get_elem(src), elem_type=elem_type, params=params, obj=dst)

        elif issubclass(elem_type, x.MessageType):
            fvalue = await self.message(x.get_elem(src), msg_type=elem_type, obj=dst)

        else:
            raise TypeError

        return x.set_elem(dst, fvalue) if not isinstance(fvalue, NoSetSentinel) else fvalue

    async def _dump_field(self, elem, elem_type, params=None, obj=None):
        return await self.field(elem=elem, elem_type=elem_type, params=params, obj=obj)

    async def _load_field(self, elem_type, params=None, elem=None, obj=None):
        return await self.field(elem=elem, elem_type=elem_type, params=params, obj=obj)


def container_is_raw(container_type, params):
    """
    Returns true if container is statically allocated array
    :param container_type:
    :param params:
    :return:
    """
    return container_type.KV_RAW_ARRAY if hasattr(container_type, 'KV_RAW_ARRAY') else False




from apps.monero.xmr.serialize.base_types import BoolType, UVarintType
from apps.monero.xmr.serialize.message_types import MessageType
from apps.monero.xmr.serialize_messages.addr import AccountPublicAddress


class TxDestinationEntry(MessageType):
    __slots__ = ("amount", "addr", "is_subaddress")

    @staticmethod
    def f_specs():
        return (
            ("amount", UVarintType),  # original: UInt64
            ("addr", AccountPublicAddress),
            ("is_subaddress", BoolType),
        )



from apps.monero.xmr.serialize.message_types import MessageType
from apps.monero.xmr.serialize_messages.base import ECKey
from apps.monero.xmr.serialize_messages.ct_keys import KeyV


class Bulletproof(MessageType):
    __slots__ = ("V", "A", "S", "T1", "T2", "taux", "mu", "L", "R", "a", "b", "t")

    @staticmethod
    def f_specs():
        return (
            ("A", ECKey),
            ("S", ECKey),
            ("T1", ECKey),
            ("T2", ECKey),
            ("taux", ECKey),
            ("mu", ECKey),
            ("L", KeyV),
            ("R", KeyV),
            ("a", ECKey),
            ("b", ECKey),
            ("t", ECKey),
        )
